from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, Collection

import yaml
from dbt.artifacts.resources import BaseResource
from dbt.artifacts.resources.v1.components import ParsedResource

from dbt_contracts.contracts.utils import get_absolute_project_path
from dbt_contracts.types import PropertiesT


class SafeLineLoader(yaml.SafeLoader):
    """YAML safe loader which applies line and column number information to every mapping read."""

    def construct_mapping(self, node: yaml.MappingNode, deep: bool = False):
        """Construct mapping object and apply line and column numbers"""
        mapping = super().construct_mapping(node, deep=deep)
        # Add 1 so line/col numbering starts at 1
        mapping["__start_line__"] = node.start_mark.line + 1
        mapping["__start_col__"] = node.start_mark.column + 1
        mapping["__end_line__"] = node.end_mark.line + 1
        mapping["__end_col__"] = node.end_mark.column + 1
        return mapping


class IndentedDumper(yaml.Dumper):
    """YAML dumper which sets extra indentation for flow collections"""

    # noinspection PyMissingOrEmptyDocstring,SpellCheckingInspection
    def increase_indent(self, flow=False, indentless=False):
        return super(IndentedDumper, self).increase_indent(flow, False)


class PropertiesIO(MutableMapping[Path, dict[str, Any]]):
    """Manages loading and saving of properties for dbt objects from/to their associated properties files"""

    def __init__(self, data: dict[Path, dict[str, Any]] = None):
        self._properties: dict[Path, dict[str, Any]] = data or {}

    def __getitem__(self, item: str | Path | PropertiesT):
        if not isinstance(item, (str, Path)):
            try:
                path = self.get_path(item=item, to_absolute=True)
                if path is None:
                    raise FileNotFoundError
            except FileNotFoundError:
                raise KeyError(f"No properties path found for the given item: {item.name!r}.")
        else:
            path = Path(item)

        if path in self._properties:
            return self._properties[path]

        if not path.suffix:
            path = path.with_suffix(".yml")

        if not path.is_file():
            raise KeyError(f"File {path} not found.")
        if path.suffix.casefold() not in {".yml", ".yaml"}:
            raise KeyError("Extension of the given properties path is not a yaml file")

        properties = self._read_file(path)
        self._properties[path] = properties
        return properties

    def __setitem__(self, key: Path, value: dict[str, Any]):
        self._properties[key] = value

    def __delitem__(self, key: Path):
        del self._properties[key]

    def __len__(self):
        return len(self._properties)

    def __iter__(self):
        return iter(self._properties)

    @staticmethod
    def _read_file(path: Path) -> dict[str, Any]:
        with path.open("r") as file:
            properties = yaml.load(file, Loader=SafeLineLoader)
        return properties

    @staticmethod
    def get_path(item: PropertiesT, to_absolute: bool = False) -> Path | None:
        """
        Get the properties path for a given item from its attributes.

        :param item: The item to get a properties path for.
        :param to_absolute: Format the path to be absolute. Only returns the path if it exists in the project.
        :return: The properties path if found.
        """
        properties_path = None
        if isinstance(item, ParsedResource) and item.patch_path:
            properties_path = Path(item.patch_path.split("://")[1])
        elif isinstance(item, BaseResource):
            if (path := Path(item.original_file_path)).suffix.casefold() in {".yml", ".yaml"}:
                properties_path = path

        if properties_path is None or not to_absolute or properties_path.is_absolute():
            return properties_path
        return get_absolute_project_path(properties_path)

    def save(self, paths: Collection[Path] = ()) -> list[Path]:
        """
        Save the stored properties.

        :param paths: Optionally, only save the properties from this list.
        """
        if not paths:
            paths = tuple(self._properties)

        paths_updated = []
        for path in paths:
            if path not in self._properties:
                continue

            properties = self._clean_properties(self._properties[path])
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("w") as file:
                yaml.dump(properties, file, Dumper=IndentedDumper, sort_keys=False, allow_unicode=True)

            paths_updated.append(path)

        return paths_updated

    @classmethod
    def _clean_properties(cls, properties: dict[str, Any]) -> dict[str, Any]:
        properties = properties.copy()

        for key, value in properties.copy().items():
            if key.startswith("__"):
                del properties[key]

            if isinstance(value, dict):
                properties[key] = cls._clean_properties(value)
            elif isinstance(value, Collection) and not isinstance(value, str):
                properties[key] = [cls._clean_properties(val) if isinstance(val, dict) else val for val in value]

        return properties



from abc import ABCMeta, abstractmethod
from copy import deepcopy
from random import choice, sample, shuffle
from unittest import mock

import pytest
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.contracts.graph.nodes import CompiledNode
from dbt_common.contracts.metadata import CatalogTable, ColumnMetadata
from faker import Faker

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.generators.column import ColumnPropertiesGenerator
from dbt_contracts.contracts.generators.node import NodePropertyGenerator, SetNodeDescription, SetNodeColumns, \
    NodePropertiesGenerator
from dbt_contracts.types import NodeT
from tests.contracts.generators.test_core import ParentPropertiesGeneratorTester
from tests.contracts.generators.test_properties import SetDescriptionTester


class TestSetNodeDescription(SetDescriptionTester[ColumnInfo, ColumnMetadata]):
    @pytest.fixture
    def generator(self) -> SetNodeDescription:
        return SetNodeDescription()

    @pytest.fixture
    def item(self, node: CompiledNode) -> CompiledNode:
        return node

    def test_run(
            self, generator: SetNodeDescription, item: CompiledNode, catalog_table: CatalogTable, faker: Faker
    ):
        item.description = faker.sentence()
        catalog_table.metadata.comment = faker.sentence()

        generator.overwrite = True

        assert generator.run(item, catalog_table)
        assert item.description == catalog_table.metadata.comment


class TestSetColumns:

    @pytest.fixture
    def generator(self) -> SetNodeColumns:
        """Fixture for the property generator to test."""
        return SetNodeColumns()

    def test_skips_on_empty_columns(self, generator: SetNodeColumns, node: CompiledNode, catalog_table: CatalogTable):
        generator.add = True
        generator.remove = True
        generator.order = True

        catalog_table = deepcopy(catalog_table)
        catalog_table.columns.clear()

        with (
                mock.patch.object(generator.__class__, "_set_column") as mock_set,
                mock.patch.object(generator.__class__, "_drop_column") as mock_drop,
                mock.patch.object(generator.__class__, "_order_columns") as mock_order,
        ):
            assert not generator.run(node, catalog_table)

            mock_set.assert_not_called()
            mock_drop.assert_not_called()
            mock_order.assert_not_called()

    def test_skips_add_when_disabled(self, generator: SetNodeColumns, node: CompiledNode, catalog_table: CatalogTable):
        assert catalog_table.columns

        with mock.patch.object(generator.__class__, "_set_column") as mock_set:
            generator.add = False
            generator.run(node, catalog_table)
            mock_set.assert_not_called()

            generator.add = True
            generator.run(node, catalog_table)
            mock_set.assert_called()

    def test_skips_remove_when_disabled(
            self, generator: SetNodeColumns, node: CompiledNode, catalog_table: CatalogTable
    ):
        assert catalog_table.columns

        with mock.patch.object(generator.__class__, "_drop_column") as mock_drop:
            generator.remove = False
            generator.run(node, catalog_table)
            mock_drop.assert_not_called()

            generator.remove = True
            generator.run(node, catalog_table)
            mock_drop.assert_called()

    def test_skips_order_when_disabled(
            self, generator: SetNodeColumns, node: CompiledNode, catalog_table: CatalogTable
    ):
        assert catalog_table.columns

        with mock.patch.object(generator.__class__, "_order_columns") as mock_order:
            generator.order = False
            generator.run(node, catalog_table)
            mock_order.assert_not_called()

            generator.order = True
            generator.run(node, catalog_table)
            mock_order.assert_called()

    def test_set_column_skips_on_matched_column(self, generator: SetNodeColumns, node: CompiledNode, faker: Faker):
        node_column = choice(list(node.columns.values()))
        column = ColumnMetadata(name=node_column.name, index=faker.random_int(), type=node_column.data_type)
        assert column.name in node.columns

        assert not generator._set_column(node, column=column)

    def test_set_column(self, generator: SetNodeColumns, node: CompiledNode, faker: Faker):
        column = ColumnMetadata(name=faker.word(), index=faker.random_int(), type="int")
        assert column.name not in node.columns

        assert generator._set_column(node, column=column)
        assert column.name in node.columns
        assert node.columns[column.name].name == column.name

    def test_drop_column_skips_on_matched_column(self, generator: SetNodeColumns, node: CompiledNode):
        columns = {
            col.name: ColumnMetadata(name=col.name, index=i, type="int")
            for i, col in enumerate(node.columns.values())
        }
        column = choice(list(node.columns.values()))
        assert column.name in columns
        assert column.name in node.columns

        assert not generator._drop_column(node, column=column, columns=columns)
        assert column.name in node.columns

    def test_drop_column(self, generator: SetNodeColumns, node: CompiledNode, faker: Faker):
        columns = {name: ColumnMetadata(name=name, index=i, type="int") for i, name in enumerate(faker.words())}
        column = choice(list(node.columns.values()))
        assert column.name not in columns
        assert column.name in node.columns

        assert generator._drop_column(node, column=column, columns=columns)
        assert column.name not in node.columns

    def test_order_columns_skips_on_empty_columns(self, generator: SetNodeColumns, node: CompiledNode):
        original_order = list(node.columns)

        assert not generator._order_columns(node, columns={})
        assert list(node.columns) == original_order

    def test_order_columns_skips_when_columns_already_in_order(
            self, generator: SetNodeColumns, node: CompiledNode, faker: Faker
    ):
        columns = {
            col.name: ColumnMetadata(name=col.name, index=faker.random_int(), type="int")
            for col in node.columns.values()
        }
        index_map = {col.name: col.index for col in columns.values()}
        node.columns = dict(
            sorted(node.columns.items(), key=lambda col: index_map.get(col[1].name, len(index_map)))
        )
        original_order = list(node.columns)

        assert not generator._order_columns(node, columns=columns)
        assert list(node.columns) == original_order

    # TODO: flakey test - fix me
    def test_order_columns(self, generator: SetNodeColumns, node: CompiledNode, faker: Faker):
        node.columns |= {name: ColumnInfo(name=name) for name in faker.words()}
        original_order = list(node.columns)
        columns = {
            col.name: ColumnMetadata(name=col.name, index=faker.random_int(max=len(node.columns)), type="int")
            for col in sample(list(node.columns.values()), k=3)
        }

        assert list(columns) != original_order

        assert generator._order_columns(node, columns=columns)
        assert list(node.columns) != original_order


class NodePropertiesGeneratorTester[I: NodeT](
    ParentPropertiesGeneratorTester[I, NodePropertyGenerator], metaclass=ABCMeta
):
    @abstractmethod
    def generator(self) -> NodePropertiesGenerator[I]:
        raise NotImplementedError

    @staticmethod
    def test_generate_column_properties(generator: ColumnPropertiesGenerator, item: I):
        column = choice(list(item.columns.values()))
        table = generator._generate_column_properties(column)
        assert all(val for val in table.values())

    # TODO: flakey test - fix me
    @staticmethod
    def test_merge_columns_merges_and_sorts(generator: NodePropertiesGenerator[I], item: I, faker: Faker):
        table = {"columns": list(map(generator._generate_column_properties, item.columns.values()))}
        modified_columns = {col["name"]: col for col in sample(table["columns"], k=3)}
        for column in modified_columns.values():
            column["description"] = faker.sentence()
            column["new_property"] = faker.random_int()

        while table["columns"] == list(modified_columns.values()):
            shuffle(table["columns"])

        expected_columns = list(map(generator._generate_column_properties, item.columns.values()))
        for column in expected_columns:
            if not (modified_column := modified_columns.get(column["name"])):
                continue
            column["new_property"] = modified_column["new_property"]

        generator._merge_columns(item, table)
        assert table["columns"] == expected_columns
        assert table["columns"] != list(modified_columns.values())  # sorted back to expected order after shuffle

    @staticmethod
    def test_merge_columns_drops(generator: NodePropertiesGenerator[I], item: I, faker: Faker):
        table = {"columns": list(map(generator._generate_column_properties, item.columns.values()))}
        added_columns = [
            generator._generate_column_properties(ColumnInfo(name=faker.word())) for _ in range(3)
        ]
        table["columns"].extend(added_columns)

        expected_columns = list(map(generator._generate_column_properties, item.columns.values()))

        generator._merge_columns(item, table)
        assert table["columns"] == expected_columns

    @staticmethod
    def test_merge_skips_on_no_table_in_database(
            generator: NodePropertiesGenerator[I], item: I, context: ContractContext
    ):
        with (
            mock.patch("dbt_contracts.contracts.generators.node.get_matching_catalog_table", return_value=None),
            mock.patch.object(
                generator.__class__, "generators", new_callable=mock.PropertyMock, return_value=[]
            ) as mock_generators,
        ):
            assert not generator.merge(item, context=context)
            mock_generators.assert_not_called()

    @staticmethod
    def test_merge(
            generator: NodePropertiesGenerator[I], item: I, context: ContractContext, catalog_table: CatalogTable
    ):
        with (
            mock.patch(
                "dbt_contracts.contracts.generators.node.get_matching_catalog_table", return_value=catalog_table
            ),
            mock.patch.object(
                generator.__class__, "generators", new_callable=mock.PropertyMock, return_value=[]
            ) as mock_generators,
            mock.patch.object(generator.description.__class__, "run", return_value=True) as mock_description,
            mock.patch.object(generator.columns.__class__, "run", return_value=False) as mock_columns,
        ):
            assert not generator.merge(item, context=context)

            mock_generators.return_value = [generator.description, generator.columns]
            assert generator.merge(item, context=context)

            mock_description.assert_called_once_with(item, catalog_table)
            mock_columns.assert_called_once_with(item, catalog_table)



from pydantic import BaseModel, ConfigDict


class BaseModelConfig(BaseModel):
    """
    Base class for all models in the dbt_contracts package.

    Sets common config and functionality as required.
    """
    model_config = ConfigDict(validate_assignment=True, validate_default=True)



from abc import ABCMeta, abstractmethod
from pathlib import Path
from unittest import mock

from dbt.artifacts.resources import BaseResource
from dbt.artifacts.resources.v1.components import ParsedResource
from dbt.flags import GLOBAL_FLAGS

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.generators import PropertyGenerator, PropertiesGenerator, ParentPropertiesGenerator, \
    ChildPropertiesGenerator
from dbt_contracts.properties import PropertiesIO
from dbt_contracts.types import ItemT, PropertiesT


class PropertiesGeneratorTester[I: ItemT, G: PropertyGenerator](metaclass=ABCMeta):
    """Base class for testing contract generators."""
    @abstractmethod
    def generator(self) -> PropertiesGenerator[I, G]:
        """Fixture for the properties generator to test."""
        raise NotImplementedError

    @abstractmethod
    def item(self, **kwargs) -> I:
        """Fixture for the item to test."""
        raise NotImplementedError

    @staticmethod
    def test_get_generators(generator: PropertiesGenerator[I, G]):
        assert not generator.exclude
        assert generator.generators
        assert all(isinstance(gen, PropertyGenerator) for gen in generator.generators)

        excluded = generator.generators[0]
        generator.exclude = [excluded.name]
        assert excluded not in generator.generators


class ParentPropertiesGeneratorTester[I: PropertiesT, G: PropertyGenerator](
    PropertiesGeneratorTester[I, G], metaclass=ABCMeta
):
    @abstractmethod
    def generator(self) -> ParentPropertiesGenerator[I, G]:
        raise NotImplementedError

    @staticmethod
    def test_update_with_no_existing_properties_path(
            generator: ParentPropertiesGenerator[I, G], item: PropertiesT, context: ContractContext, tmp_path: Path
    ):
        GLOBAL_FLAGS.PROJECT_DIR = tmp_path

        item.original_file_path = ""
        item.patch_path = None
        assert not context.properties.get_path(item)

        expected_path = tmp_path.joinpath(generator.generate_properties_path(item))
        expected_path.parent.mkdir(parents=True)
        expected_path.touch()

        with (
            mock.patch.object(generator.__class__, "_update_existing_properties") as mock_update,
            mock.patch.object(generator.__class__, "_generate_properties") as mock_generate,
            mock.patch.object(PropertiesIO, "__getitem__", return_value=None) as mock_get,
            mock.patch.object(PropertiesIO, "__setitem__") as mock_set,
        ):
            generator.update(item, context=context)

            mock_update.assert_not_called()
            mock_generate.assert_called_once()
            mock_get.assert_called_once()
            mock_set.assert_called_once()

            properties_path = context.properties.get_path(item, to_absolute=True)
            assert properties_path is not None
            assert tmp_path.joinpath(properties_path) in context.properties

            relative_path = properties_path.relative_to(tmp_path)
            if isinstance(item, ParsedResource):
                assert item.patch_path == f"{context.manifest.metadata.project_name}://{relative_path}"
            elif isinstance(item, BaseResource):
                assert item.original_file_path == str(relative_path)

    @staticmethod
    def test_update_with_existing_properties_path(
            generator: ParentPropertiesGenerator[I, G], item: PropertiesT, context: ContractContext, tmp_path: Path
    ):
        assert context.properties.get_path(item)

        GLOBAL_FLAGS.PROJECT_DIR = tmp_path
        expected_path = tmp_path.joinpath(context.properties.get_path(item))
        expected_path.parent.mkdir(parents=True)
        expected_path.touch()

        with (
            mock.patch.object(generator.__class__, "_update_existing_properties") as mock_update,
            mock.patch.object(generator.__class__, "_generate_properties") as mock_generate,
            mock.patch.object(PropertiesIO, "__getitem__", return_value=None) as mock_get,
            mock.patch.object(PropertiesIO, "__setitem__") as mock_set,
        ):
            generator.update(item, context=context)

            mock_update.assert_called_once()
            mock_generate.assert_not_called()
            mock_get.assert_called_once()
            mock_set.assert_not_called()

    @staticmethod
    def test_update_with_existing_properties_file(
            generator: ParentPropertiesGenerator[I, G], item: PropertiesT, context: ContractContext
    ):
        properties = {"key": "value"}
        item.original_file_path = ""
        item.patch_path = None
        assert not context.properties.get_path(item)

        with (
            mock.patch.object(generator.__class__, "_update_existing_properties") as mock_update,
            mock.patch.object(generator.__class__, "_generate_properties") as mock_generate,
            mock.patch.object(PropertiesIO, "__getitem__", return_value=properties) as mock_get,
            mock.patch.object(PropertiesIO, "__setitem__") as mock_set,
        ):
            generator.update(item, context=context)

            mock_update.assert_called_once()
            mock_generate.assert_not_called()
            mock_get.assert_called_once()
            mock_set.assert_not_called()

    @staticmethod
    def test_generate_properties_path_with_no_set_depth(
            generator: ParentPropertiesGenerator[I, G], item: PropertiesT, tmp_path: Path
    ):
        GLOBAL_FLAGS.PROJECT_DIR = tmp_path
        assert generator.depth is None
        generator.filename = "props"  # no extension given, extension suffix will be added

        expected = tmp_path.joinpath("path", "to", "a", "different", f"{generator.filename}.yml")
        item.original_file_path = expected.relative_to(tmp_path).with_name(Path(item.path).name)
        with mock.patch.object(PropertiesIO, "get_path", return_value=None):
            assert generator.generate_properties_path(item) == expected

    @staticmethod
    def test_generates_properties_path_with_depth(
            generator: ParentPropertiesGenerator[I, G], item: PropertiesT, tmp_path: Path
    ):
        GLOBAL_FLAGS.PROJECT_DIR = tmp_path
        item.path = str(tmp_path.joinpath("path", "to", "a", "model"))
        generator.depth = 1  # takes parents up to index=1
        generator.filename = "properties.yaml"  # valid extension given, suffix will be kept as is

        expected = tmp_path.joinpath("path", "to", generator.filename)
        item.path = expected.relative_to(tmp_path).with_name(Path(item.path).name)
        item.original_file_path = None
        with mock.patch.object(PropertiesIO, "get_path", return_value=None):
            assert generator.generate_properties_path(item) == expected


class ChildPropertiesGeneratorTester[I: ItemT, P: PropertiesT, G](PropertiesGeneratorTester[I, G], metaclass=ABCMeta):
    @abstractmethod
    def generator(self) -> ChildPropertiesGenerator[I, P, G]:
        raise NotImplementedError

    @abstractmethod
    def parent(self, **kwargs) -> P:
        """Fixture for the parent item to test."""
        raise NotImplementedError



from unittest import mock

from dbt.contracts.graph.nodes import ModelNode
from dbt_common.contracts.constraints import ModelLevelConstraint, ConstraintType

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.terms.model import HasConstraints


def test_has_constraints(model: ModelNode, context: ContractContext):
    model.constraints.extend((
        ModelLevelConstraint(type=ConstraintType.not_null, columns=["col1", "col2"]),
        ModelLevelConstraint(type=ConstraintType.unique, columns=["col1", "col2"]),
    ))
    assert HasConstraints().run(model, context=context)

    with mock.patch.object(ContractContext, "add_result") as mock_add_result:
        assert not HasConstraints(min_count=5).run(model, context=context)
        mock_add_result.assert_called_once()



from typing import Any

from dbt.contracts.graph.nodes import SourceDefinition

from dbt_contracts.contracts.generators.node import NodePropertiesGenerator
from dbt_contracts.contracts.utils import merge_maps


class SourcePropertiesGenerator(NodePropertiesGenerator):

    def _update_existing_properties(self, item: SourceDefinition, properties: dict[str, Any]) -> dict[str, Any]:
        key = item.resource_type.pluralize()
        if key not in properties:
            properties[key] = []

        source = next((source for source in properties[key] if source["name"] == item.source_name), None)
        if source is None:
            source = self._generate_source_properties(item)
            properties[key].append(source)
        if "tables" not in source:
            source["tables"] = []

        table_in_props = next((prop for prop in source["tables"] if prop["name"] == item.name), None)
        table = self._generate_table_properties(item)
        if table_in_props is not None:
            merge_maps(table_in_props, table, overwrite=True, extend=False)
            table = table_in_props
        else:
            source["tables"].append(table)

        self._merge_columns(item, table)

        return properties

    def _generate_properties(self, item: SourceDefinition) -> dict[str, Any]:
        key = item.resource_type.pluralize()
        columns = list(map(self._generate_column_properties, item.columns.values()))
        table = self._generate_table_properties(item) | {"columns": columns}
        source = self._generate_source_properties(item) | {"tables": [table]}

        return self._properties_defaults | {key: [source]}

    @staticmethod
    def _generate_source_properties(item: SourceDefinition) -> dict[str, Any]:
        source = {
            "name": item.source_name,
            "description": item.source_description,
            "database": item.unrendered_database or item.database,
            "schema": item.unrendered_schema or item.schema,
        }
        return {key: val for key, val in source.items() if val}

    @staticmethod
    def _generate_table_properties(item: SourceDefinition) -> dict[str, Any]:
        table = {
            "name": item.name,
            "description": item.description,
        }
        return {key: val for key, val in table.items() if val}



from unittest import mock

import pytest
from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt.contracts.graph.nodes import ModelNode

from dbt_contracts.contracts import ContractContext
from dbt_contracts.contracts.result import Result, RESULT_PROCESSOR_MAP
from dbt_contracts.types import ItemT, ParentT


class TestContractContext:
    def test_context_init(self):
        context = ContractContext()
        assert len(context.results) == 0
        assert context.properties is ContractContext.properties

        with pytest.raises(TypeError):  # cannot manually set results
            # noinspection PyArgumentList
            ContractContext(results=None)

    @staticmethod
    def assert_result(results: list[Result], name: str, message: str, item: ItemT, parent: ParentT = None):
        """Assert that a result with the given name and message exists in the results list"""
        assert any(result.name == item.name for result in results)
        assert any(result.result_name == name for result in results)
        assert any(result.message == message for result in results)

        if parent is None:
            return

        assert any(result.parent_id == parent.unique_id for result in results)
        assert any(result.parent_name == parent.name for result in results)

    def test_add_result_on_item(self, context: ContractContext, model: ModelNode):
        expected_name = "test_name"
        expected_message = "this test has failed"

        cls = RESULT_PROCESSOR_MAP[type(model)]
        with mock.patch.object(cls, "_extract_properties_for_item", return_value={}):
            context.add_result(item=model, name=expected_name, message=expected_message)

        self.assert_result(context.results, item=model, name=expected_name, message=expected_message)

    def test_add_result_on_item_with_parent(self, context: ContractContext, model: ModelNode, column: ColumnInfo):
        expected_name = "test_name"
        expected_message = "this test has failed"
        context.add_result(item=column, parent=model, name=expected_name, message=expected_message)

        self.assert_result(context.results, item=column, parent=model, name=expected_name, message=expected_message)



from pathlib import Path
from unittest import mock

import pytest
from colorama import Fore

from dbt_contracts.contracts.result import ModelResult, Result
from dbt_contracts.formatters.table import TableCellBuilder, TableRowBuilder, TableFormatter, GroupedTableFormatter


class TestTableCellBuilder:
    @pytest.fixture
    def result(self, tmp_path: Path) -> Result:
        """A fixture for a result object."""
        return ModelResult(
            name="this is a result",
            path=tmp_path,
            result_type="failure",
            result_level="error",
            result_name="result",
            message="This is an error message.",
        )

    def test_validate_wrap(self):
        TableCellBuilder(key="name", wrap=True, max_width=30)  # valid
        with pytest.raises(Exception):
            TableCellBuilder(key="name", wrap=True)

    def test_prefix_coloured(self):
        assert TableCellBuilder(key="name").prefix_coloured == ""
        assert TableCellBuilder(key="name", prefix="pre-").prefix_coloured == "pre-"

        result = TableCellBuilder(key="name", prefix="pre-", colour=Fore.CYAN).prefix_coloured
        expected = f"{Fore.CYAN.replace("m", ";1m")}pre-{Fore.RESET.replace("m", ";0m")}"
        assert result == expected

    def test_get_value(self, result: Result):
        builder = TableCellBuilder(key="name")
        assert builder._get_value(result) == "this is a result"

        builder = TableCellBuilder(key=lambda x: f"{x.name} ({x.result_type})")
        assert builder._get_value(result) == "this is a result (failure)"

    def test_apply_prefix(self):
        builder = TableCellBuilder(key="name")
        assert builder._apply_prefix("value") == "value"

        builder = TableCellBuilder(key="name", prefix="pre-")
        assert builder._apply_prefix("") == ""
        assert builder._apply_prefix(f"{Fore.CYAN}{Fore.RESET}") == f"{Fore.CYAN}{Fore.RESET}"
        assert builder._apply_prefix("value") == "pre-value"

        builder = TableCellBuilder(key="name", prefix="pre-", colour=Fore.CYAN)
        assert builder._apply_prefix("value") == f"{builder.prefix_coloured}value" != "pre-value"

    # noinspection SpellCheckingInspection
    def test_truncate_value(self):
        builder = TableCellBuilder(key="name", max_width=5)
        assert builder._truncate_value("value") == "value"

        builder = TableCellBuilder(key="name", max_width=5)
        assert builder._truncate_value("value") == "value"

        builder = TableCellBuilder(key="name", max_width=5)
        assert builder._truncate_value("valuevalue") == "valu…"

    def test_apply_padding_and_alignment(self):
        builder = TableCellBuilder(key="name")
        assert builder._apply_padding_and_alignment("value", width=10) == f"{"value":<10}"

        builder = TableCellBuilder(key="name", alignment=">")
        assert builder._apply_padding_and_alignment("value", width=10) == f"{"value":>10}"

        builder = TableCellBuilder(key="name", alignment="^")
        assert builder._apply_padding_and_alignment("value", width=10) == f"{"value":^10}"

    def test_apply_colour(self):
        builder = TableCellBuilder(key="name")
        assert builder._apply_colour("value") == "value"

        builder = TableCellBuilder(key="name", colour=Fore.CYAN)
        assert builder._apply_colour("value") == f"{Fore.CYAN}value{Fore.RESET}"

    def test_apply_wrap(self):
        builder = TableCellBuilder(key="name")
        value = "i am a very long value"
        assert builder._apply_wrap(value) == [value]

        builder = TableCellBuilder(key="name", max_width=10, wrap=True)
        assert builder._apply_wrap(value) == ["i am a", "very long", "value"]

    def test_build(self, result: Result):
        builder = TableCellBuilder(key="name")
        assert builder.build(result) == "this is a result"

        builder = TableCellBuilder(key="name", prefix="pre-", min_width=20, max_width=14, colour=Fore.CYAN)
        assert builder.build(result) == f"{builder.prefix_coloured}{Fore.CYAN}this is a…{Fore.RESET}      "

        builder = TableCellBuilder(
            key="name", prefix="pre-", alignment=">", min_width=15, max_width=10, wrap=True, colour=Fore.RED
        )
        expected_width = builder.min_width + len(Fore.RED + Fore.RESET)

        expected_width_with_prefix = expected_width - len(builder.prefix)
        expected = (
            f"{builder.prefix_coloured}{f"{builder.colour}this{Fore.RESET}":>{expected_width_with_prefix}}\n"
            f"{f"{builder.colour}is a{Fore.RESET}":>{expected_width}}\n"
            f"{f"{builder.colour}result{Fore.RESET}":>{expected_width}}"
        )
        assert builder.build(result) == expected

    def test_build_uses_min_width(self, result: Result):
        builder = TableCellBuilder(key="name", min_width=20, max_width=10)
        assert builder.build(result) == f"{"this is a…":<20}"
        assert builder.build(result, min_width=40) == f"{"this is a…":<40}"


class TestTableRowBuilder:
    @pytest.fixture
    def result(self, tmp_path: Path) -> Result:
        """A fixture for a result object."""
        return ModelResult(
            name="this is a result",
            path=tmp_path,
            result_type="failure",
            result_level="error",
            result_name="result",
            message="This is an error message.",
        )

    @pytest.fixture
    def cells(self) -> list[TableCellBuilder]:
        """A fixture for a table cell builders."""
        return [
            TableCellBuilder(key="name", min_width=10, max_width=20),
            TableCellBuilder(key="result_type", min_width=10, max_width=20),
            TableCellBuilder(key="result_level", min_width=10, max_width=20),
        ]

    @pytest.fixture
    def builder(self, cells: list[TableCellBuilder]) -> TableRowBuilder:
        """Fixture for a table row builder."""
        return TableRowBuilder(cells=cells)

    def test_validate_cells(self, cells: list[TableCellBuilder]):
        cells_row_2 = cells.copy()
        # noinspection PyTypeChecker
        cells_row_2[0] = None

        # all valid
        TableRowBuilder(cells=cells)
        TableRowBuilder(cells=[cells])
        TableRowBuilder(cells=[cells, cells_row_2])

        with pytest.raises(Exception):
            TableRowBuilder(cells=cells_row_2)
        with pytest.raises(Exception):
            TableRowBuilder(cells=[cells_row_2])
        with pytest.raises(Exception):
            TableRowBuilder(cells=[cells, cells[:-1]])

    def test_separator_coloured(self, cells: list[TableCellBuilder]):
        assert TableRowBuilder(cells=cells).separator_coloured == "|"
        assert TableRowBuilder(cells=cells, separator=",").separator_coloured == ","

        builder = TableRowBuilder(cells=cells, separator=",", colour=Fore.RED)
        expected = f"{Fore.RED.replace("m", ";1m")},{Fore.RESET.replace("m", ";0m")}"
        assert builder.separator_coloured == expected

    def test_get_lines_simple(self, result: Result, cells: list[TableCellBuilder]):
        min_widths = [None] * len(cells)

        builder = TableRowBuilder(cells=cells)
        assert builder._get_lines(result, min_widths) == ["this is a result", "failure   ", "error     "]
        builder = TableRowBuilder(cells=[cells])
        assert builder._get_lines(result, min_widths) == ["this is a result", "failure   ", "error     "]

    def test_get_lines_multi_row(self, result: Result, cells: list[TableCellBuilder]):
        cells = [cells, [None, TableCellBuilder(key="message"), None]]
        min_widths = [None] * len(cells[0])
        expected = ["this is a result\n", "failure   \nThis is an error message.", "error     \n"]

        builder = TableRowBuilder(cells=cells)
        assert builder._get_lines(result, min_widths) == expected

    def test_get_lines_with_min_widths(self, result: Result, cells: list[TableCellBuilder]):
        cells = [cells, [None, TableCellBuilder(key="message"), None]]
        min_widths = [20, 30, 0]
        expected = [
            "this is a result    \n",
            "failure                       \nThis is an error message.     ",
            "error\n"
        ]

        builder = TableRowBuilder(cells=cells)
        assert builder._get_lines(result, min_widths) == expected

    def test_get_lines_with_wrap(self, result: Result, cells: list[TableCellBuilder]):
        cells[0].max_width = 5
        cells[0].wrap = True
        cells = [cells, [TableCellBuilder(key="result_name"), TableCellBuilder(key="message"), None]]
        min_widths = [None] * len(cells[0])
        expected = [
            "this      \nis a      \nresult    \nresult    ",
            "failure   \nThis is an error message.",
            "error     \n"
        ]

        builder = TableRowBuilder(cells=cells)
        assert builder._get_lines(result, min_widths) == expected

    def test_get_max_length(self):
        lines = ["this is a cell", "this is another cell", "this is the last cell"]
        assert TableRowBuilder._get_max_rows(lines) == 1

        lines[2] = "this is\nthe last\ncell"
        assert TableRowBuilder._get_max_rows(lines) == 3

        lines[1] = "this is\nanother cell"
        assert TableRowBuilder._get_max_rows(lines) == 3

        lines[0] = "this\nis\na\ncell"
        assert TableRowBuilder._get_max_rows(lines) == 4

    def test_to_matrix(self):
        lines = ["this is a cell", "this is another cell", "this is the last cell"]
        result = TableRowBuilder._to_matrix(lines)
        assert result == [tuple(lines)]

        lines[1] = "this is\nanother cell"
        result = TableRowBuilder._to_matrix(lines)
        assert result == [
            ("this is a cell", "this is", "this is the last cell"),
            ("", "another cell", ""),
        ]

        lines[2] = "this\nis\nthe\nlast cell"
        result = TableRowBuilder._to_matrix(lines)
        assert result == [
            ("this is a cell", "this is", "this"),
            ("", "another cell", "is"),
            ("", "", "the"),
            ("", "", "last cell"),
        ]

        lines[0] = "this is\na\ncell"
        result = TableRowBuilder._to_matrix(lines)
        assert result == [
            ("this is", "this is", "this"),
            ("a", "another cell", "is"),
            ("cell", "", "the"),
            ("", "", "last cell"),
        ]

    def test_remove_empty_lines(self):
        lines = [["this is a cell", "this is another cell"]]
        assert TableRowBuilder._remove_empty_lines(lines) == lines

        lines.append(["this is a 2nd row cell", "         "])
        assert TableRowBuilder._remove_empty_lines(lines) == lines

        lines.append(["            ", "         "])
        assert TableRowBuilder._remove_empty_lines(lines) == lines[:2]

    def test_get_widths_from_lines(self):
        lines = [["this is a cell", "this is another cell", "this is the last cell"]]
        assert TableRowBuilder.get_widths_from_lines(lines) == [len(lines[0][0]), len(lines[0][1]), len(lines[0][2])]

        lines = [
            ["this is", "this is", "this"],
            ["a", "another cell", "is"],
            ["cell", " " * len("another cell"), "the"],
            [" " * len("this is"), " " * len("another cell"), "last cell"],
        ]
        assert TableRowBuilder.get_widths_from_lines(lines) == [7, 12, 9]

    def test_extend_line_widths_fails(self, builder: TableRowBuilder, result: Result):
        # noinspection PyTypeChecker
        min_widths = [None] * (len(builder.cells[0]) - 1)
        with pytest.raises(Exception):  # too few min widths given
            builder.extend_line_widths([], min_widths=min_widths)

    def test_extend_line_widths(self, builder: TableRowBuilder, result: Result):
        lines = [[
            ["this is a result", "failure", "error", "keep as is"],
            ["result", "This is an error message.", "", ""],
            ["this is a result", "failure", "error", ""],
        ]]
        builder.cells[0][2].alignment = ">"
        builder.cells[0].append(None)
        min_widths = [20, 30, 10, None]
        expected = [[
            ["this is a result    ", "failure                       ", "     error", "keep as is"],
            ["result              ", "This is an error message.     ", "          ", ""],
            ["this is a result    ", "failure                       ", "     error", ""],
        ]]

        assert builder.extend_line_widths(lines, min_widths=min_widths) == expected

    def test_build_lines_fails(self, builder: TableRowBuilder, result: Result):
        builder.build_lines(result)  # valid because no min_widths given

        # noinspection PyTypeChecker
        min_widths = [None] * (len(builder.cells[0]) - 1)
        with pytest.raises(Exception):  # too few min widths given
            builder.build_lines(result, min_widths=min_widths)

    def test_build_lines(self, builder: TableRowBuilder, result: Result):
        assert builder.build_lines(result) == [["this is a result", "failure   ", "error     "]]

        builder.cells = [
            builder.cells[0],
            [TableCellBuilder(key="result_name"), TableCellBuilder(key="message"), None],
            [  # produces an empty line
                TableCellBuilder(key="properties_path"),
                TableCellBuilder(key="properties_start_line"),
                TableCellBuilder(key="properties_start_col")
            ],
            builder.cells[0],
        ]
        expected = [
            ["this is a result", "failure                  ", "error     "],
            ["result          ", "This is an error message.", "          "],
            ["this is a result", "failure                  ", "error     "],
        ]
        assert builder.build_lines(result) == expected

    def test_build_lines_uses_min_widths(self, builder: TableRowBuilder, result: Result):
        builder.cells = [
            builder.cells[0],
            [TableCellBuilder(key="result_name"), TableCellBuilder(key="message"), None],
            [  # produces an empty line
                TableCellBuilder(key="properties_path"),
                TableCellBuilder(key="properties_start_line"),
                TableCellBuilder(key="properties_start_col")
            ],
            builder.cells[0],
        ]
        min_widths = [20, 30, 0]
        expected = [
            ["this is a result    ", "failure                       ", "error"],
            ["result              ", "This is an error message.     ", "     "],
            ["this is a result    ", "failure                       ", "error"],
        ]
        assert builder.build_lines(result, min_widths=min_widths) == expected

    def test_build_lines_pads_cell_lines(self, builder: TableRowBuilder, result: Result):
        builder.cells = [
            builder.cells[0],
            [TableCellBuilder(key="result_name"), TableCellBuilder(key="message", max_width=5, wrap=True), None],
            builder.cells[0],
        ]
        expected = [
            ["this is a result", "failure   ", "error     "],
            ["result          ", "This      ", "          "],
            ["this is a result", "is an     ", "error     "],
            ["                ", "error     ", "          "],
            ["                ", "message.  ", "          "],
            ["                ", "failure   ", "          "]
        ]
        assert builder.build_lines(result) == expected

    def test_join(self, builder: TableRowBuilder, result: Result):
        builder.cells[0].append(TableCellBuilder(key="message", min_width=10, max_width=6, wrap=True))
        builder.colour = Fore.RED
        sep = builder.separator_coloured
        lines = [[
            ["this is a result", "failure   ", "error     ", "This      "],
            ["                ", "          ", "          ", "is an     "],
            ["                ", "          ", "          ", "error     "],
            ["                ", "          ", "          ", "message.  "]
        ], [
            ["this is result 2", "success   ", "pass      ", "This      "],
            ["                ", "          ", "          ", "is a      "],
            ["                ", "          ", "          ", "success   "],
            ["                ", "          ", "          ", "message.  "]
        ]]

        assert builder.join(lines) == "\n".join((
            f"this is a result {sep} failure    {sep} error      {sep} This      ",
            f"                 {sep}            {sep}            {sep} is an     ",
            f"                 {sep}            {sep}            {sep} error     ",
            f"                 {sep}            {sep}            {sep} message.  ",
            f"this is result 2 {sep} success    {sep} pass       {sep} This      ",
            f"                 {sep}            {sep}            {sep} is a      ",
            f"                 {sep}            {sep}            {sep} success   ",
            f"                 {sep}            {sep}            {sep} message.  ",
        ))

    def test_build(self, builder: TableRowBuilder, result: Result):
        builder.separator = ":"
        assert builder.build(result) == "this is a result : failure    : error     "

        builder.cells[0].append(TableCellBuilder(key="message", min_width=10, max_width=6, wrap=True))
        builder.colour = Fore.RED
        sep = builder.separator_coloured
        assert builder.build(result) == "\n".join((
            f"this is a result {sep} failure    {sep} error      {sep} This      ",
            f"                 {sep}            {sep}            {sep} is an     ",
            f"                 {sep}            {sep}            {sep} error     ",
            f"                 {sep}            {sep}            {sep} message.  ",
        ))

    def test_build_uses_min_widths(self, builder: TableRowBuilder, result: Result):
        min_widths = [1, 2, 3]
        with mock.patch.object(TableRowBuilder, "build_lines", return_value=[]) as mock_build_lines:
            builder.build(result, min_widths=min_widths)
            mock_build_lines.assert_called_once_with(result, min_widths=min_widths)


class TestTableBuilder:
    @pytest.fixture
    def results(self, tmp_path: Path) -> list[Result]:
        """A fixture for the results."""
        return [
            ModelResult(
                name="this is the 1st result",
                path=tmp_path,
                result_type="failure",
                result_level="error",
                result_name="result",
                message="This is an error message.",
            ),
            ModelResult(
                name="this is the 2nd result",
                path=tmp_path,
                result_type="a very great success",
                result_level="info",
                result_name="result",
                message="This is a success message.",
            )
        ]

    @pytest.fixture
    def formatter(self, ) -> TableFormatter:
        """A fixture for the table formatter."""
        cells = [
            TableCellBuilder(key="name", min_width=10, max_width=20),
            TableCellBuilder(key="result_type", min_width=10, max_width=20),
            TableCellBuilder(key="result_level", min_width=10, max_width=20),
        ]
        row = TableRowBuilder(cells=cells)
        return TableFormatter(builder=row)

    def test_add_header(self, formatter: TableFormatter):
        formatter.add_header("i am a header")
        assert formatter._lines == ["i am a header", ""]

        formatter.add_header("i am a 2nd header")
        assert formatter._lines == ["i am a 2nd header", ""]

        formatter._lines = ["i am a line", "i am another line"]
        formatter.add_header("i am a 3rd header")
        assert formatter._lines == ["i am a 3rd header", "", "i am a line", "i am another line"]

    def test_add_results(self, results: list[Result], formatter: TableFormatter):
        formatter.consistent_widths = False
        formatter.add_results(results)
        assert formatter._results == [
            [["this is the 1st res…", "failure   ", "error     "]],
            [["this is the 2nd res…", "a very great success", "info      "]],
        ]

    def test_add_results_with_consistent_widths(self, results: list[Result], formatter: TableFormatter):
        formatter.consistent_widths = True
        formatter.add_results(results)
        assert formatter._results == [
            [["this is the 1st res…", "failure             ", "error     "]],
            [["this is the 2nd res…", "a very great success", "info      "]],
        ]

    def test_build(self, results: list[Result], formatter: TableFormatter):
        formatter.consistent_widths = True
        formatter.add_results(results)
        assert formatter.build() == (
            "this is the 1st res… | failure              | error     \n"
            "this is the 2nd res… | a very great success | info      "
        )


class TestGroupedTableFormatter:
    @pytest.fixture
    def results(self, tmp_path: Path) -> list[Result]:
        """A fixture for the results."""
        return [
            ModelResult(
                name="this is the 1st result",
                path=tmp_path,
                result_type="failure",
                result_level="error",
                result_name="result",
                message="This is an error message.",
            ),
            ModelResult(
                name="this is the 5th result",
                path=tmp_path,
                result_type="warning",
                result_level="warning",
                result_name="result",
                message="This is a warning message.",
            ),
            ModelResult(
                name="this is the 3rd result",
                path=tmp_path,
                result_type="another very great success",
                result_level="info",
                result_name="result",
                message="This is a success message.",
            ),
            ModelResult(
                name="this is the 4th result",
                path=tmp_path,
                result_type="failure",
                result_level="error",
                result_name="result",
                message="This is an error message.",
            ),
            ModelResult(
                name="this is the 2nd result",
                path=tmp_path,
                result_type="a very great success",
                result_level="info",
                result_name="result",
                message="This is a success message.",
            ),
        ]

    @pytest.fixture
    def formatter(self, ) -> GroupedTableFormatter:
        """A fixture for the grouped table formatter."""
        cells = [
            TableCellBuilder(key="name", min_width=10, max_width=20),
            TableCellBuilder(key="result_type", min_width=10, max_width=20),
            TableCellBuilder(key="result_level", min_width=10, max_width=20),
        ]
        row = TableRowBuilder(cells=cells)
        table = TableFormatter(builder=row)
        return GroupedTableFormatter(formatter=table, group_key="result_level")

    def test_get_value(self, results: list[Result]):
        assert GroupedTableFormatter._get_value(results[0], getter="name") == "this is the 1st result"

        assert GroupedTableFormatter._get_value(
            results[0], getter=lambda x: f"{x.name} ({x.result_type})"
        ) == "this is the 1st result (failure)"

    def test_add_results_basic(self, formatter: GroupedTableFormatter, results: list[Result]):
        assert not formatter.header_key
        assert not formatter.sort_key

        with (
                mock.patch.object(TableFormatter, "add_header") as mock_header,
                mock.patch.object(TableFormatter, "add_results") as mock_results,
                mock.patch.object(TableFormatter, "build") as mock_build,
        ):
            formatter.add_results(results)

            assert len(mock_header.mock_calls) == 3
            assert len(mock_results.mock_calls) == 3
            assert len(mock_build.mock_calls) == 3

            group_keys = {result.result_level for result in results}
            for key in group_keys:
                mock_header.assert_any_call(key)
                group = [result for result in results if result.result_level == key]
                mock_results.assert_any_call(group)

    def test_add_results_with_header(self, formatter: GroupedTableFormatter, results: list[Result]):
        assert not formatter.sort_key
        formatter.header_key = "message"

        with (
                mock.patch.object(TableFormatter, "add_header") as mock_header,
                mock.patch.object(TableFormatter, "add_results") as mock_results,
                mock.patch.object(TableFormatter, "build") as mock_build,
        ):
            formatter.add_results(results)

            assert len(mock_header.mock_calls) == 3
            assert len(mock_results.mock_calls) == 3
            assert len(mock_build.mock_calls) == 3

            group_headers = {result.result_level: result.message for result in results}
            for key, message in group_headers.items():
                mock_header.assert_any_call(message)
                group = [result for result in results if result.result_level == key]
                mock_results.assert_any_call(group)

    def test_add_results_with_sort(self, formatter: GroupedTableFormatter, results: list[Result]):
        assert not formatter.header_key
        formatter.sort_key = ["name"]

        with (
                mock.patch.object(TableFormatter, "add_header") as mock_header,
                mock.patch.object(TableFormatter, "add_results") as mock_results,
                mock.patch.object(TableFormatter, "build") as mock_build,
        ):
            formatter.add_results(results)

            assert len(mock_header.mock_calls) == 3
            assert len(mock_results.mock_calls) == 3
            assert len(mock_build.mock_calls) == 3

            group_keys = {result.result_level for result in results}
            for key in group_keys:
                group = [result for result in results if result.result_level == key]
                group = sorted(group, key=lambda r: r.name)
                mock_results.assert_any_call(group)

    def test_build(self, formatter: GroupedTableFormatter, results: list[Result]):
        assert not formatter._tables
        assert not formatter.build()

        formatter.add_results(results)
        assert formatter._tables

        assert formatter.build()
        assert not formatter._tables



from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod

from dbt_contracts.contracts._contracts import ContractPart
from dbt_contracts.types import ItemT


class ContractCondition[T: ItemT](ContractPart, metaclass=ABCMeta):
    """
    Conditional logic to apply to items within the manifest to determine
    whether they should be processed by subsequent terms.
    """
    @classmethod
    def _name(cls) -> str:
        """The name of this condition in snake_case."""
        class_name = cls.__name__.replace("Condition", "")
        return re.sub(r"([a-z])([A-Z])", r"\1_\2", class_name).lower()

    @abstractmethod
    def run(self, item: T) -> bool:
        """Run this condition to check whether the given item should be processed."""
        raise NotImplementedError



from abc import ABCMeta
from collections.abc import Sequence
from random import choice, sample
from typing import Literal, Annotated, get_args

from dbt.artifacts.resources.v1.components import ColumnInfo
from dbt_common.contracts.metadata import ColumnMetadata
from pydantic import Field, BeforeValidator

from dbt_contracts.contracts._core import ContractContext
from dbt_contracts.contracts.generators._core import ChildPropertiesGenerator, CORE_FIELDS, PropertyGenerator
from dbt_contracts.contracts.generators.properties import SetDescription
from dbt_contracts.contracts.utils import get_matching_catalog_table, to_tuple
from dbt_contracts.types import NodeT

COLUMN_FIELDS = Literal[CORE_FIELDS, "data_type"]


class ColumnPropertyGenerator(PropertyGenerator[ColumnInfo, ColumnMetadata], metaclass=ABCMeta):
    pass


class SetColumnDescription(ColumnPropertyGenerator, SetDescription[ColumnInfo, ColumnMetadata]):
    def run(self, source: ColumnInfo, target: ColumnMetadata) -> bool:
        return self._set_description(source, description=target.comment)


class SetDataType(ColumnPropertyGenerator):
    @classmethod
    def _name(cls) -> str:
        return "data_type"

    def _set_data_type(self, source: ColumnInfo, data_type: str | None) -> bool:
        if not data_type:
            return False
        if source.data_type and not self.overwrite:
            return False
        if source.data_type == data_type:
            return False

        source.data_type = data_type
        return True

    def run(self, source: ColumnInfo, target: ColumnMetadata) -> bool:
        return self._set_data_type(source, data_type=target.type)


EXCLUDE_TYPES = Literal["description", "data_type"]


class ColumnPropertiesGenerator[P: NodeT](ChildPropertiesGenerator[ColumnInfo, P, ColumnPropertyGenerator]):
    exclude: Annotated[Sequence[EXCLUDE_TYPES], BeforeValidator(to_tuple)] = Field(
        description="The fields to exclude from the generated properties.",
        default=(),
        examples=[choice(get_args(EXCLUDE_TYPES)), sample(get_args(EXCLUDE_TYPES), k=2)]
    )
    description: SetColumnDescription = Field(
        description="Configuration for setting the column description",
        default=SetColumnDescription(),
    )
    data_type: SetDataType = Field(
        description="Configuration for setting the column data type",
        default=SetDataType(),
    )

    def merge(self, item: ColumnInfo, context: ContractContext, parent: P = None) -> bool:
        if (table := get_matching_catalog_table(parent, catalog=context.catalog)) is None:
            return False
        if (column := next((col for col in table.columns.values() if col.name == item.name), None)) is None:
            return False

        return any([generator.run(item, column) for generator in self.generators])
